/* Diana.java
 *
 * created: Fri Oct  9 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/Diana.java,v 1.74 1999/08/02 10:00:14 kmr Exp $
 */

import diana.components.*;
import diana.sequence.*;
import diana.*;

import uk.ac.sanger.pathogens.*;

import uk.ac.sanger.pathogens.embl.QualifierInfoVector;
import uk.ac.sanger.pathogens.embl.QualifierInfoException;
import uk.ac.sanger.pathogens.embl.EntryInformation;
import uk.ac.sanger.pathogens.embl.ReadFormatException;
import uk.ac.sanger.pathogens.embl.InvalidRelationException;
import uk.ac.sanger.pathogens.embl.LocationParseException;
import uk.ac.sanger.pathogens.embl.InvalidKeyException;

import nsdb.EmblSeq;
import type.NoResult;
import nsdb.EmblPackage.Superceded;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;

/**
 *  This class contains the method main () for running Diana as a standalone
 *  program.  This class is also used by DianaApplet as a storage place for
 *  the global Clipboard object and the global Options object.
 *
 *  @author Kim Rutherford
 *  @version $Id: Diana.java,v 1.74 1999/08/02 10:00:14 kmr Exp $
 * */

public class Diana extends Frame {
  /**
   *  Main entry point for the stand-alone Diana.
   **/
  public static void main (final String [] args) {
    db_ior_url = getOptions ().getProperty ("db_ior_url");

    embl_ior_url = getOptions ().getProperty ("embl_ior_url");

    try {
      final StringVector extra_keys =
        getOptions ().getOptionValues ("extra_keys");

      final QualifierInfoVector extra_qualifiers =
        getOptions ().getExtraQualifiers ();

      final Document feature_keys_document =
        new URLDocument (Diana.class.getResource ("/etc/feature_keys"));

      final Document qualifier_document =
        new URLDocument (Diana.class.getResource ("/etc/qualifier_types"));

      EntryInformation.initialise (feature_keys_document,
                                   qualifier_document,
                                   extra_keys,
                                   extra_qualifiers);
    } catch (IOException e) {
      System.out.println ("could not initialise the embl package: " +
                          e.getMessage ());
      System.exit (1);
    } catch (QualifierInfoException e) {
      System.out.println ("could not initialise the embl package: " +
                          e.getMessage ());
      System.exit (1);
    }

    main_window = new Diana ();

    main_window.setVisible (true);

    // read the entries given on the command line and in the diana.ini file
    main_window.readArgsAndOptions (args);
  }

  /**
   *  Read the entries given on the command line and in the diana.ini file.
   **/
  private void readArgsAndOptions (final String [] args) {
    if (args.length == 0) {
      // open the entries given in the options file (diana.ini)
      readDefaultEntries ();
    }

    EntryEdit last_entry_edit = null;

    for (int i = 0 ; i < args.length ; ++i) {
      final String new_entry_name = args[i];

//        System.out.println ("token " + new_entry_name);

      if (new_entry_name.startsWith ("+") && last_entry_edit != null) {
//          System.out.println ("adding " + new_entry_name);

        final uk.ac.sanger.pathogens.embl.Entry new_embl_entry =
          openEntry (new_entry_name.substring (1));

        if (new_entry_name == null) {
          // the read failed
          break;
        }

        try {
          final Entry new_entry =
            new Entry (last_entry_edit.getEntryGroup ().getBases (),
                       new_embl_entry);

          last_entry_edit.getEntryGroup ().add (new_entry);
        } catch (OutOfRangeException e) {
          new MessageDialog (this, "read failed: one of the features in " +
                             new_entry_name + " has an out of range location");
        }

      } else {
        final uk.ac.sanger.pathogens.embl.Entry new_embl_entry =
          openEntry (new_entry_name);

        if (new_embl_entry == null) {
          // the read failed
          break;
        }

        final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

        final EntryGroup entry_group = new EntryGroup (embl_bases);

        try {
          final Entry new_entry =
            new Entry (entry_group.getBases (), new_embl_entry);

          entry_group.add (new_entry);

          status_line.setText ("");

          last_entry_edit = makeEntryEdit (entry_group);
        } catch (OutOfRangeException e) {
          new MessageDialog (this, "read failed: one of the features in " +
                             new_entry_name + " has an out of range location");
          break;
        }
      }
    }

    for (int entry_index = 0 ;
         entry_index < entry_edit_objects.size () ;
         ++entry_index) {
      entry_edit_objects.elementAt (entry_index).show ();
    }
  }

  /**
   *  Read the entries given in the diana.ini file.
   **/
  private void readDefaultEntries () {
    final String default_sequence_file_name =
      getOptions ().getDefaultSequenceFileName ();

    final String default_feature_file_name =
      getOptions ().getDefaultFeatureFileName ();

    if (default_sequence_file_name != null) {
      final String default_sequence_file_name_embl =
        default_sequence_file_name + "_embl";

      uk.ac.sanger.pathogens.embl.Entry new_embl_entry = null;

      // try opening the default sequence file with "_embl" added to the name
      // if that fails try the plain sequence file name
      if (new File (default_sequence_file_name_embl).exists ()) {
        new_embl_entry = openEntry (default_sequence_file_name_embl);
      }

      if (new_embl_entry == null) {
        new_embl_entry = openEntry (default_sequence_file_name);
      }

      if (new_embl_entry == null) {
        // read failed
        return;
      }

      status_line.setText ("");

      final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

      final EntryGroup entry_group = new EntryGroup (embl_bases);

      try {
        final Entry new_entry =
          new Entry (entry_group.getBases (), new_embl_entry);

        entry_group.add (new_entry);

        final EntryEdit new_entry_edit = makeEntryEdit (entry_group);

        if (default_feature_file_name != null) {
          final uk.ac.sanger.pathogens.embl.Entry new_embl_table_entry =
            openEntry (default_feature_file_name);

          if (new_embl_table_entry == null) {
            // open failed
            return;
          }

          final Entry new_table_entry =
            new Entry (entry_group.getBases (), new_embl_table_entry);

          entry_group.add (new_table_entry);
        }
      } catch (OutOfRangeException e) {
        // quietly ignore this entry
      }
    }
  }

  /**
   *  Open the Entry with the given file name or show a message if there is no
   *  document with that name.
   *  @param new_entry_name The file name to open.
   *  @return The reference of the new embl.Entry or null if the read failed.
   **/
  private uk.ac.sanger.pathogens.embl.Entry openEntry (final String
                                                         new_entry_name) {
    try {
      System.out.println ("trying to open: " + new_entry_name);

      final Document new_document =
        new FileProgressDocument (new File (new_entry_name), status_line);

//      final Entry new_entry = Entry.openEntry (entry_group, new_document);

      if (! new_document.readable ()) {
        // need to create a new Document before continuing
        new MessageDialog (this, "cannot open this file: "  + new_entry_name);
        return null;
      }

      final uk.ac.sanger.pathogens.embl.Entry new_entry =
        new uk.ac.sanger.pathogens.embl.DocumentEntry (new_document);

//    System.out.println ("Entry.openEntry (): " + entry);

      return new_entry;
    } catch (ReadFormatException e) {
      final String message =
        "failed to open " + new_entry_name + " - " +
        e.getMessage () + " at line: " + e.getLineNumber ();
      System.out.println (message);
      new MessageDialog (this, message);
    } catch (IOException e) {
      final String message =
        "failed to open: " + e.getMessage ();
      status_line.setText (message);
      new MessageDialog (this, message);
    } catch (InvalidRelationException e) {
      final String message =
        "illegal qualifier: " + e.getMessage ();
      status_line.setText (message);
      new MessageDialog (this, message);
    }

    return null;
  }

  /**
   *  The constructor creates all the components for the main Diana window and
   *  sets up all the menu callbacks.
   **/
  private Diana () {
    super ("Artemis");

    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        exit ();
      }
    });

    setLayout(new BorderLayout());

    final Font font = getOptions ().getFont ();

    setFont (font);

    makeAllMenus ();

    final Canvas helix_canvas = makeHelixCanvas ();

    add (helix_canvas, "Center");

    add (status_line, "South");

    pack ();

    setSize (450, 200);

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));
  }

  /**
   *  This method will fill the background with white.
   **/
  public void paint (final Graphics g) {
    g.setColor (Color.white);
    g.fillRect (0, 0, getSize ().width, getSize ().height);
  }

  /**
   *  Return a Canvas object that will display a helix and a short copyright
   *  notice.
   **/
  private Canvas makeHelixCanvas () {
    return new Canvas () {
      public void update(Graphics g) {
        paint(g);
      }

      public void paint (final Graphics g) {
        if (helix == null) {
          Toolkit toolkit = Toolkit.getDefaultToolkit ();

          final URL helix_url = Diana.class.getResource ("/diana/helix.gif");
          helix = toolkit.getImage (helix_url);

          final URL sanger_url =
            Diana.class.getResource ("/diana/sanger-centre.gif");
          sanger = toolkit.getImage (sanger_url);

          tracker = new MediaTracker (this);

          tracker.addImage (helix, 0);
          tracker.addImage (sanger, 1);

          try {
            tracker.waitForAll ();
            helix_height = helix.getHeight (this);
            helix_width = helix.getWidth (this);
            sanger_height = sanger.getHeight (this);
          } catch (InterruptedException e) {
            return;
          }
        }

        final Font canvas_font = new Font ("Monospaced", Font.PLAIN, 16);

        g.setFont (canvas_font);

        FontMetrics fm = getFontMetrics (canvas_font);

        final int font_height = fm.getHeight () + 3;

        g.setColor (Color.white);
        g.fillRect (0, 0, getSize ().width, getSize ().height);

        for (int i = 0 ; i * helix_height <= getSize ().height; ++i) {
          g.drawImage (helix,
                       0,
                       i * helix_height, this);
        }

        g.drawImage (sanger,
                     helix_width + 5,
                     getSize ().height - sanger_height, this);

        g.setColor (Color.black);


        g.drawString ("Artemis",
                      helix_width + 5, font_height);
        g.drawString ("Copyright 1998,1999",
                      helix_width + 5, font_height * 2);
        g.drawString ("Genome Research Limited",
                      helix_width + 5, font_height * 3);
      }

      Image helix = null;
      Image sanger = null;

      MediaTracker tracker = null;

      int sanger_height;
      int helix_height;
      int helix_width;
    };
  }

  /**
   *  This method gets rid of an EntryEdit object and it's frame.  Each object
   *  removed with this method must have been added previously with
   *  addEntryEdit ().
   *  @param entry_edit The object to get rid of.
   **/
  static public synchronized void entryEditFinished (EntryEdit entry_edit) {
    if (null == entry_edit) {
      throw new Error ("Diana.entryEditFinished () was passed a null object");
    }

    if (!entry_edit_objects.removeElement (entry_edit)) {
      throw new Error ("Diana.entryEditFinished () - could not remove a " +
                       "object from an empty vector");
    }
    entry_edit.setVisible (false);
    entry_edit.dispose ();
  }

  /**
   *  Add an EntryEdit object to our list of objects.
   *  @param entry_edit The object to add.
   **/
  static public synchronized void addEntryEdit (EntryEdit entry_edit) {
    entry_edit_objects.addElement (entry_edit);
  }

  /**
   *  Return the reference of the global clipboard object for this program.
   **/
  static public ClipBoard getClipboard () {
    return clipboard;
  }

  /**
   *  Return the reference of the global options object for this program.
   **/
  static public Options getOptions () {
    return Options.getOptions ();
  }

  /**
   *  Return true if and only if we are running in an applet.
   **/
  public static boolean isApplet () {
    return is_applet;
  }

  /**
   *  Set whether we are running as an applet.
   **/
  static void setIsApplet (boolean is_applet) {
    Diana.is_applet = is_applet;
  }

  /**
   *  Return true if and only if we are not running in an applet.
   **/
  public static boolean isStandAlone () {
    if (isApplet ()) {
      return false;
    } else {
      return true;
    }
  }

  /**
   *  Make a new menu item in the given menu, with its label given the
   *  String and add the given ActionListener to it.
   */
  private static void makeMenuItem (Menu menu, String name,
                                    ActionListener listener) {
    MenuItem new_item = new MenuItem (name);
    menu.add (new_item);
    new_item.addActionListener (listener);
  }

  /**
   *  Make all the menus and menu items for the main window.  Also sets up
   *  suitable ActionListener objects for each item.
   */
  private void makeAllMenus () {

    setMenuBar (menu_bar);

    menu_bar.add (file_menu);
//    menu_bar.add (edit_menu);
    menu_bar.add (options_menu);

    ActionListener menu_listener;

    menu_listener = new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        makeOpenDialog ();
      }
    };
    makeMenuItem (file_menu, "Open ...", menu_listener);

    if (db_ior_url != null) {
      menu_listener = new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          makeCorbaDialog (null, db_ior_url, false);
        }
      };
      makeMenuItem (file_menu, "Open From Database ...", menu_listener);
    }

    if (embl_ior_url != null) {
      menu_listener = new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          makeCorbaDialog (null, embl_ior_url, true);
        }
      };
      makeMenuItem (file_menu, "Read from EBI ...", menu_listener);
    }

//      menu_listener = new ActionListener () {
//        public void actionPerformed (ActionEvent event) {
//          makeCorbaDialog (sanger_ior_url, true);
//        }
//      };
//      makeMenuItem (file_menu, "Read from Sanger ...", menu_listener);

    menu_listener = new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        exit ();
      }
    };
    makeMenuItem (file_menu, "Quit", menu_listener);

    menu_listener = new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        resetOptions ();
      }
    };
    makeMenuItem (options_menu, "Re-read Options", menu_listener);

    final Options options = Options.getOptions ();

    final CheckboxMenuItem enable_direct_edit_item =
      new CheckboxMenuItem ("Enable Direct Editing");
    enable_direct_edit_item.setState (options.canDirectEdit ());
    enable_direct_edit_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        options.setDirectEdit (enable_direct_edit_item.getState ());
      }
    });
    options_menu.add (enable_direct_edit_item);
  }

  /**
   *  Force the options files to be re-read and the EntryEdit components to be
   *  redisplayed.
   **/
  private void resetOptions () {
    Options.getOptions ().reset ();
  }

  /**
   *  Force all the EntryEdit components to be redisplayed.
   **/
  private void redisplayAll () {
    for (int i = 0 ; i < entry_edit_objects.size () ; ++i) {
      entry_edit_objects.elementAt (i).redisplay ();
    }
  }

  /**
   *  Close the main Diana frame and all EntryEdit frames and then this frame,
   *  then exit.
   **/
  private void exit () {
    for (int i = 0 ; i < entry_edit_objects.size () ; ++i) {
      entryEditFinished (entry_edit_objects.elementAt (i));
    }
    setVisible (false);
    dispose ();
    System.gc ();
    System.exit (0);
  }

  /**
   *  Create a file dialog and if the user selects a file then call
   *  makeEntryEdit () with the entry.
   **/
  private EntryEdit makeOpenDialog () {
    final EntryFileDialog dialog = new EntryFileDialog (main_window);

    final uk.ac.sanger.pathogens.embl.Entry new_embl_entry =
      dialog.getEntry ();

    if (new_embl_entry != null) {

      status_line.setText ("");

      final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

      final EntryGroup entry_group = new EntryGroup (embl_bases);

      try {
        final Entry new_entry =
          new Entry (entry_group.getBases (), new_embl_entry);

        entry_group.add (new_entry);

        final EntryEdit entry_edit = makeEntryEdit (entry_group);

        entry_edit.setVisible (true);

        return entry_edit;
      } catch (OutOfRangeException e) {
        new MessageDialog (this, "read failed: one of the features in " +
                           new_embl_entry.getName () +
                           " has an out of range location");
      }
    }

    return null;
  }

  /**
   *  Create a TextRequester, wait for the user to type an accession number
   *  and then read that entry from the server.
   *  @param entry_group If this is null a new EntryEdit component will be
   *    created for the entry once it has been read from the server.  If not
   *    null then the new entry will be added to the given EntryGroup.
   *  @param ior_url The URL of the IOR for the corba server.
   *  @param read_only true if and only if the corba server should be opened
   *    read-only (some are always read only).
   **/
  public static void makeCorbaDialog (final EntryGroup entry_group,
                                      final String ior_url_string,
                                      final boolean read_only) {
    final org.omg.CORBA.ORB orb =
      org.omg.CORBA.ORB.init (new String [0], new java.util.Properties());

    final org.omg.CORBA.Object obj;

    try {
      obj = orb.string_to_object (getIOR (ior_url_string));
    } catch (IOException e) {
      final String message =
        "Error while reading EMBL IOR: " + e;
      status_line.setText (message);
      new MessageFrame (message);
      return;
    }

    final nsdb.Embl corba_embl;

    try {
      corba_embl = nsdb.EmblHelper.narrow (obj);
    } catch (org.omg.CORBA.COMM_FAILURE e) {
      final String message = "Failed to get an object from Corba/EMBL: " + e;
      status_line.setText (message);
      new MessageFrame (message);
      return;
    }

    if (corba_embl == null) {
      final String message = "Failed to get an object from Corba/EMBL";
      status_line.setText (message);
      new MessageFrame (message);
      return;
    }

    System.out.println ("reading from " + ior_url_string);

    final TextRequester text_requester =
      new TextRequester ("Enter an accession number:", 10, "");

    final TextRequesterListener listener =
      new TextRequesterListener () {
        public void actionPerformed (final TextRequesterEvent event) {
          if (event.getType () == TextRequesterEvent.CANCEL) {
            return;
          }

          final String corba_id = event.getRequesterText ().trim ();

          if (corba_id.length () > 0 ) {
            final MessageFrame message_frame =
            new MessageFrame ("reading entry - please wait");
            makeFromCorbaID (entry_group, corba_embl, corba_id, read_only);
            message_frame.dispose ();
          }
        }
      };

    text_requester.addTextRequesterListener (listener);

    text_requester.setVisible (true);
  }

  /**
   *  Given an accession number the handle of an EMBL corba server, this
   *  method will ask the user (using a TextRequester) for the id of a entry
   *  in the server and will then attempt to get it.
   *  @param entry_group If this is null a new EntryEdit component will be
   *    created for the entry once it has been read from the server.  If not
   *    null then the new entry will be added to the given EntryGroup.
   *  @param corba_handle The handle of the nsdb.Embl object from which we
   *    will read the entry.
   *  @param corba_id The id of the entry in the database
   *  @param read_only true if and only if the corba server should be opened
   *    read-only (some are always read only).
   **/
  private static void makeFromCorbaID (final EntryGroup entry_group,
                                       final nsdb.Embl corba_handle,
                                       final String corba_id,
                                       final boolean read_only) {
    try {
      final uk.ac.sanger.pathogens.embl.Entry new_embl_entry;

      if (read_only) {
        final EmblSeq embl_seq = corba_handle.getEmblSeq (corba_id);

        new_embl_entry = new uk.ac.sanger.pathogens.embl.CorbaEntry (embl_seq);
      } else {
        nsdb.EmblWriter embl_writer =
          nsdb.EmblWriterHelper.narrow (corba_handle);

        final nsdb.EmblSeqWriter embl_write_seq =
          embl_writer.getEmblSeqWriter (corba_id);

        new_embl_entry =
          new uk.ac.sanger.pathogens.embl.RWCorbaEntry (embl_write_seq);
      }

      final EntryGroup real_entry_group;

      if (entry_group == null) {
        if (new_embl_entry.getSequence () == null) {
          final String message =
            "the entry contains no sequence: " + corba_id;
          status_line.setText (message);
          new MessageFrame (message);
          return;
        }

        final Bases embl_bases = new Bases (new_embl_entry.getSequence ());

        real_entry_group = new EntryGroup (embl_bases);
      } else {
        real_entry_group = entry_group;
      }

      final Entry new_entry =
        new Entry (real_entry_group.getBases (), new_embl_entry);

      real_entry_group.add (new_entry);

      status_line.setText ("");

      if (entry_group == null) {
        final EntryEdit entry_edit = makeEntryEdit (real_entry_group);

        entry_edit.setVisible (true);
      }
    } catch (NoResult e) {
      final String message =
        "Database query failed while getting id: " + corba_id;
      status_line.setText (message);
      new MessageFrame (message);
    } catch (Superceded e) {
      //  Superceded is used by getEmblSeq method if accession number
      //  doesn't exist anymore because it was merged or split
      final String message =
        "This accession number has been superceded: " + corba_id;
      status_line.setText (message);
      new MessageFrame (message);
    } catch (LocationParseException e) {
      final String message =
        "Unexpected error while accessing " + corba_id + ": " + e;
      status_line.setText (message);
      new MessageFrame (message);
    } catch (InvalidKeyException e) {
      final String message =
        "Unexpected error while accessing " + corba_id + ": " + e;
      status_line.setText (message);
      new MessageFrame (message);
    } catch (org.omg.CORBA.OBJECT_NOT_EXIST e) {
      final String message =
        "the object you requested (" + corba_id + ") does not exist";
      status_line.setText (message);
      new MessageFrame (message);
    } catch (java.lang.Throwable e) {
      final String message =
        "Unexpected error while accessing " + corba_id + ": " + e;
      status_line.setText (message);
      new MessageFrame (message);

      e.printStackTrace ();
    }
  }

  /**
   *  Finds and returns a stringified object reference (IOR).
   *  @param ior_url_string The URL of the IOR for the corba server.
   *  @return The stringified IOR.
   *  @exception IOException Thrown if an error occurs while reading.
   **/
  private static String getIOR (final String ior_url_string)
      throws IOException {
    try {
      //  get URL of IOR
      URL url = new URL (ior_url_string);
      BufferedReader in =
        new BufferedReader (new InputStreamReader (url.openStream()));
      return in.readLine();
    } catch (MalformedURLException e) {
      throw new Error ("internal error - unexpected MalformedURLException: " +
                       e.getMessage ());
    }
  }

  /**
   *  Create a EntryEdit object from an EMBL file and then call
   *  addEntryEdit ().
   *  @param entry The embl.Entry object for the new EntryEdit object.
   *  @return The newly created EntryEdit component.
   **/
  public static EntryEdit makeEntryEdit (EntryGroup entry_group) {

    final EntryEdit entry_edit = new EntryEdit (entry_group);

    addEntryEdit (entry_edit);

    return entry_edit;
  }

  /**
   *  The location of the IOR for the corba server at EMBL.
   **/
  static private String embl_ior_url =
    "http://corba.ebi.ac.uk/EMBL/IOR/Embl.IOR";

  /**
   *  The location of the IOR for the corba server at The Sanger Centre.
   **/
  static private final String sanger_ior_url =
    "http://localhost/~kmr/ior";

  /**
   *  The location of the IOR for the read-write corba server.
   **/
  static private String db_ior_url;

  /**
   *  The main diana frame.
   **/
  static private Diana main_window;

  /**
   *  Menu bar for the main window.
   **/
  static final private MenuBar menu_bar = new MenuBar ();

  /**
   *  The file menu.
   **/
  static final private Menu file_menu = new Menu ("File");

  /**
   *   The edit menu.
   **/
  static final private Menu edit_menu = new Menu ("Edit");

  /**
   *   The options menu.
   **/
  static final private Menu options_menu = new Menu ("Options");

  /**
   *
   **/
  static final private Panel panel = new Panel ();

  /**
   *  A label for status and error messages.
   **/
  static final private Label status_line = new Label ("");

  /**
   *  A vector containing all the EntryEdit object we have created.
   **/
  static private EntryEditVector entry_edit_objects = new EntryEditVector ();

  /**
   *  The global clip board object.  Used by the stand alone program and the
   *  applet.
   **/
  static private ClipBoard clipboard = new ClipBoard ();

  /**
   *  This is true if and only if we are running as an applet.
   **/
  static private boolean is_applet = false;
}
