/* SelectionViewer.java
 *
 * created: Thu Mar  4 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/SelectionViewer.java,v 1.8 1999/06/11 10:05:28 kmr Exp $
 **/

package diana.components;

import diana.*;
import diana.sequence.*;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

/**
 *  This component displays the current selection in a FileViewer component.
 *
 *  @author Kim Rutherford
 *  @version $Id: SelectionViewer.java,v 1.8 1999/06/11 10:05:28 kmr Exp $
 **/

public class SelectionViewer
    implements SelectionChangeListener {
  /**
   *  Create a new SelectionViewer object from the given Selection.
   **/
  public SelectionViewer (final Selection selection,
                          final EntryGroup entry_group) {
    file_viewer = new FileViewer ("Artemis Selection View");

    this.selection = selection;
    this.entry_group = entry_group;

    readSelection ();

    selection.addSelectionChangeListener (this);

    file_viewer.close_button.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        stopListening ();
      }
    });

    file_viewer.addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        stopListening ();
      }
    });
  }

  /**
   *  Remove this object as a selection change listener.
   **/
  public void stopListening () {
    selection.removeSelectionChangeListener (this);
  }

 /**
   *  Implementation of the SelectionChangeListener interface.  We listen to
   *  SelectionChange events so that we can update the view to reflect the
   *  current selection.
   **/
  public void selectionChanged (SelectionChangeEvent event) {
    readSelection ();
  }

  /**
   *  Read the selection into this SelectionViewer object.
   **/
  public void readSelection () {
    final String base_selection_text =
      SelectionInfoDisplay.markerRangeText (selection, entry_group);

    final FeatureVector selection_features = selection.getAllFeatures ();

    final StringBuffer buffer = new StringBuffer ();

    // this is the maximum number of bases to show
    final int display_base_count = 2000;

    if (base_selection_text != null && base_selection_text.length () > 0) {

      buffer.append (base_selection_text).append ("\n\n");

      final MarkerRange marker_range = selection.getMarkerRange ();

      final String selection_bases = Strand.markerRangeBases (marker_range);

      if (selection_bases.length () > 0) {
        int gc_count = 0;

        for (int i = 0 ; i < selection_bases.length () ; ++i) {
          final char this_char = selection_bases.charAt (i);
          if (this_char == 'g' || this_char == 'c') {
            ++gc_count;
          }
        }
      
        buffer.append ("gc percentage of the selected bases: " +
                       10000 * gc_count / selection_bases.length () / 100.0 +
                       "\n\n");
      }

      if (selection_bases.length () > display_base_count) {
        // display the first display_base_count bases and the last
        // display_base_count bases
        buffer.append ("first " + display_base_count / 2 +
                       " bases of selection:\n");
        final String start_sub_string =
          selection_bases.substring (0, display_base_count / 2);
        buffer.append (start_sub_string).append ("\n");
        buffer.append ("\n").append ("last " + display_base_count / 2 +
                                     " bases of selection:\n");
        final String sub_string =
          selection_bases.substring (selection_bases.length () -
                                         display_base_count / 2);
        buffer.append (sub_string).append ("\n");
      } else {
        buffer.append ("\n").append ("bases of selection:\n");
        buffer.append (selection_bases);
      }
    }

    for (int i = 0 ; i < selection_features.size () && i < 50 ; ++i) {
      buffer.append (selection_features.elementAt (i).toString ());
    }

    file_viewer.readString (buffer.toString ());
  }

  /**
   *  This is the selection object that we are viewing.
   **/
  final private Selection selection;

  /**
   *  The FileViewer object that is displaying the selection.
   **/
  final private FileViewer file_viewer;

  /**
   *  The EntryGroup that was passed to the constructor.
   **/
  final EntryGroup entry_group;
}


