/* FeatureGCAlgorithm.java
 *
 * created: Wed Dec 16 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/plot/FeatureGCAlgorithm.java,v 1.4 1999/05/11 15:19:13 kmr Exp $
 */

package diana.plot;

import diana.*;
import uk.ac.sanger.pathogens.embl.Range;
import diana.sequence.Bases;

/**
 *  Objects of this class have one useful method - getValues (), which takes a
 *  range of bases and returns a single floating point number, which is the
 *  percent GC content of the range.
 *
 *  @author Kim Rutherford
 *  @version $Id: FeatureGCAlgorithm.java,v 1.4 1999/05/11 15:19:13 kmr Exp $
 **/

public class FeatureGCAlgorithm extends FeatureAlgorithm {
  /**
   *  Create a new FeatureGCAlgorithm object.
   *  @param feature The feature to do future calculation on.
   **/
  public FeatureGCAlgorithm (Feature feature) {
    super (feature, "GC Content (%)");
  }

  /**
   *  Return the hydrophobicity between a pair of amino acids.
   *  @param start The start amino acid (included in the range).
   *  @param end The end amino acid (included in the range).
   *  @param values The one return value for this algorithm is returned in
   *    this array.
   **/
  public void getValues (int start, int end, float [] values) {
    final String sequence =
      getFeature ().getBases ().substring (start, end);

//    System.out.println (start + " " + end);

    float gc_count = 0;

    for (int i = 0 ; i < sequence.length () ; ++i) {
      final char this_char = sequence.charAt (i);

      if (this_char == 'g' || this_char == 'c') {
        ++gc_count;
      }
    }

//    System.out.println ("start: " + start + " end: " + end + " returning: " + gc_count/sequence.length ()); 

    values[0] = gc_count/sequence.length () * 100;
  }

  /**
   *  Return the number of values a call to getValues () will return - three
   *  in this case.
   **/
  public int getValueCount () {
    return 1;
  }

  /**
   *  Return the default or optimal step size.
   *  @return null is returned if this algorithm doesn't have optimal step
   *    size.
   **/
  public Integer getDefaultStepSize (int window_size) {
    // must step by three for these algorithms because they work on codons
    return new Integer (5);
  } 

  /**
   *  Return the average value of function over the whole strand.
   *  @return null is returned if this algorithm doesn't have an average or if
   *    the average can't be calculated.
   **/
  public Float getAverage () {
    final float average =
      getFeature ().getStrand ().getBases ().getAverageGCPercent ();
    return new Float (average);
  } 

}


